import { AsyncSeriesWaterfallHook } from 'tapable';
import { isFn } from '@spark-build/web-utils/lib/type';
import { PluginApi } from './Api';
import { loadPlugins } from './util';
import { ApplyPluginsType } from '../../constants';
export class ServerPlugin {
  constructor(server) {
    this.plugins = new Map();
    this.skipIds = new Set();
    this.initPluginReady = false;
    this.server = server;
  }

  get isReady() {
    return this.initPluginReady;
  }

  getPlugins() {
    return this.plugins;
  }

  register(id, plugin) {
    this.plugins.set(id, plugin);
    return this;
  }

  skip(id) {
    this.skipIds.add(id);
  }

  removeSkip(id) {
    if (this.skipIds.has(id)) {
      this.skipIds.delete(id);
    }
  }

  enableStatus(id) {
    if (!this.plugins.has(id) || this.skipIds.has(id)) {
      return false;
    }

    if (this.server.userConfig[id] === false) {
      return false;
    }

    return true;
  }

  getPluginAPI(id) {
    const plugin = new PluginApi(id, this.server);
    return new Proxy(plugin, {
      get: (target, prop) => {
        if (this.server.methods.has(prop)) {
          return this.server.methods.get(prop);
        }

        const findProp = (t = target) => isFn(t[prop]) ? t[prop].bind(t) : t[prop];

        if (this[prop]) {
          return findProp(this);
        }

        if (this.server[prop]) {
          return findProp(this.server);
        }

        return findProp();
      }
    });
  }

  async initPlugins() {
    const plugins = await loadPlugins();
    const tInitPlugin = new AsyncSeriesWaterfallHook(['initPlugin']);

    for (const plugin of plugins) {
      tInitPlugin.tapPromise({
        name: plugin.id,
        stage: plugin.stage
      }, async () => {
        const pluginInstance = this.getPluginAPI(plugin.id);
        await plugin.fn(pluginInstance);
        return this.register(plugin.id, pluginInstance);
      });
    }

    await tInitPlugin.promise(undefined);
    this.initPluginReady = true;
    this.server.applyPlugins({
      key: 'onPluginReady',
      type: ApplyPluginsType.EVENT
    });
    return this;
  }

}